
/* Symmetricom Time & Frequency Processor Library
 *
 * This library is an Sgi Irix user-level driver (pciba) for the BC635pci Time &
 * Frequency Processor board from Symmetricom.  It was developed on Irix
 * 6.5.22.
 *
 * On Irix user-level interrupts (ULI) is a feature of an optional product
 * called REACT/pro.  You will need to purchase a license from Sgi.
 *
 * ---
 *
 * This code is based on code supplied by Symmetricom that I was told was
 * originnaly written by another customer.  The original code was interspersed
 * within the menu based utility.  This library has been separated from that
 * program into this file and heavily rewritten and expanded.  I wouldn't call
 * it polished and definitely not fully tested but it seems to mostly work.
 *
 * The ULI part is pretty flakey.  I originally couldn't get ULI to work.  I
 * eventually tried an example program from the Sgi documentation that was for some
 * other timerboard.  I made the minimum modifications, using code from here, to
 * get it to receive interrupts.  It worked. So then I stripped out all of the code
 * I wasn't really using from the example.  It failed too. Then I started over
 * striping out a little bit of the useless code at a time.  What I found is that
 * if you make all of the calls as if you are going to use the cycle counter then
 * ULI works and you can receive interrupts.  I do not know why, since we aren't
 * using the cycle counter.  I incorporated the cycle counter code into this
 * library to get ULI interrupt reception working.  So, if you are looking at the
 * code in the #if CYCLE_COUNTER sections and wondering why it is there, now you
 * know.
 *
 * David M. Cooke
 * Software Enginner, Unisys Corporation
 * Assigned to FSSB at NASA LaRC
 * d.m.cooke@larc.nasa.gov
 *
 * ---
 *
 * */

#define CYCLE_COUNTER 1

#include <stdio.h>
#include <stdlib.h>
#include <fcntl.h>
#include <sched.h>
#include <time.h>
#include <unistd.h>
#include <sys/types.h>
#include <sys/mman.h>
#include <sys/fcntl.h>
#include <sys/prctl.h>
#include <sys/syssgi.h>
#include <sys/sysmp.h>
#include <sys/time.h>
#include <sys/uli.h>

#include <sys/PCI/PCI_defs.h>
#include <sys/PCI/pciba.h>

#include "stfplib.h"


#ifndef TRUE
#define TRUE 1
#endif
#ifndef FALSE
#define FALSE 0
#endif

#define  BC_VENDOR_ID        0x12e2         /* PCI Vendor ID for bc635 */
#define  BC635_DEVICE_ID     0x4013         /* PCI Device ID for bc635 */

#define INTERRUPT_STACK_SIZE 16384

struct stfp_registers
{
  uint32_t  timereq;			/* time request (for time[] regs) */
  uint32_t  eventreq;			/* time request (for event[] regs) */
  uint32_t  unlock;			/* time capture lockout release */
  uint32_t  reserved;
  uint32_t  control;			/* misc. control register */
  uint32_t  ack;			/* acknowledge register */
  uint32_t  mask;			/* interrupt mask register */
  uint32_t  intstat;			/* interrupt status register */
  uint32_t  minstrb;			/* minor strobe time */
  uint32_t  majstrb;			/* major strobe time */
  uint32_t  reserved2[2];
  uint32_t  time[2];			/* minor/major time hold */
  uint32_t  event[2];			/* minor/major event */
};

struct stfp_device
{
  char*                            config_name;
  char*                            mem_name;
  char*                            intr_name;
  int                              config_handle;   /* config device handle returned from open() */
  int                              mem_handle;      /* mem device handle returned from open() */
  int                              intr_handle;     /* intr device handle returned from open() */
  void*                            ULI_id;          /* ULI id returned from ULI_register_pci() */
  void*                            interrupt_stack; /* stack for interrupt function */
  uint32_t                         pci_id;          /* Vendor ID in botton 16 bits,
                                                     *  device ID in top 16 bits
                                                     */
  uint8_t                          pci_revision;    /* PCI revision number */
  uint8_t                          pci_header;      /* PCI header type */
  uint32_t                         bar[6];          /* Base address registers */
  volatile struct stfp_registers*  registers;       /* Device control registers (BAR0) */
  volatile uint8_t*                dpram;           /* Dual-port RAM (BAR1) */
  volatile uint8_t*                input_area;      /* DPRAM Input Area (data from user) */
  volatile uint8_t*                output_area;     /* DPRAM Output Area (data to user) */
  volatile uint8_t*                gps_area;        /* DPRAM GPS packet area */
  volatile uint8_t*                year_area;       /* DPRAM year area */
#if CYCLE_COUNTER
  int                              mmem_fd;         /* for cycle counter */
  volatile unsigned long long*     timer_address;   /* cycle counter */
#endif
};

const int  dpram_size = 0x800;

static uint32_t btoh32(uint32_t board)
{
  uint32_t  host;
  char*     host_bytes  = (char*)&host;
  char*     board_bytes = (char*)&board;

  host_bytes[0] = board_bytes[3];
  host_bytes[1] = board_bytes[2];
  host_bytes[2] = board_bytes[1];
  host_bytes[3] = board_bytes[0];

  return host;
}


static uint32_t htob32(uint32_t host)
{
  uint32_t  board;
  char*     board_bytes = (char*)&board;
  char*     host_bytes  = (char*)&host;

  board_bytes[0] = host_bytes[3];
  board_bytes[1] = host_bytes[2];
  board_bytes[2] = host_bytes[1];
  board_bytes[3] = host_bytes[0];

  return board;
}

#if CYCLE_COUNTER
static void stfp_open_cycle_counter(struct stfp_device*  stfp)
{
  int  poffmask;
  unsigned int  cycleval;
  __psunsigned_t  phys_addr;
  __psunsigned_t  raddr;

  poffmask = getpagesize() - 1;
  phys_addr = syssgi(SGI_QUERY_CYCLECNTR, &cycleval);
  raddr = phys_addr & ~poffmask;
  stfp->mmem_fd = open("/dev/mmem", O_RDONLY);
  stfp->timer_address = (volatile unsigned long long *)mmap(0,
                                                            poffmask,
                                                            PROT_READ,
                                                            MAP_PRIVATE,
                                                            stfp->mmem_fd,
                                                            (off_t)raddr);
  stfp->timer_address = (unsigned long long*)((__psunsigned_t)stfp->timer_address +
                                              (phys_addr & poffmask));

}
#endif

uint32_t  stfp_get_interrupt_enable_mask(struct stfp_device*  stfp)
{
  return stfp->registers->mask;
}


void  stfp_set_interrupt_enable_mask(struct stfp_device*  stfp, uint32_t  mask)
{
  mask &= (STFP_INTERRUPT_EVENT_INPUT |
           STFP_INTERRUPT_PERODIC_OUTPUT |
           STFP_INTERRUPT_TIME_COINCIDENCE_STROBE |
           STFP_INTERRUPT_1_PPS_OUTPUT |
           STFP_INTERRUPT_GPS_DATA_PACKET_AVAILABLE |
           STFP_INTERRUPT_EVENT_IN_LOGIC_LEVEL |
           STFP_INTERRUPT_DCLS_IN_LOGIC_LEVEL |
           STFP_INTERRUPT_EXTERNAL_1_PPS_IN_LOGIC_LEVEL);

  stfp->registers->mask = mask;
}


uint32_t  stfp_get_interrupt_status(struct stfp_device*  stfp)
{
  return stfp->registers->intstat;
}


void  stfp_clear_interrupt_status(struct stfp_device*  stfp, uint32_t  mask)
{
  stfp->registers->intstat = mask;
}


void stfp_time_request(struct stfp_device*  stfp)
{
  stfp->registers->timereq = 0;
}


void stfp_get_time_registers(struct stfp_device*  stfp, uint32_t*  time_registers)
{
  if (!time_registers)
    return;

  time_registers[0] = btoh32(stfp->registers->time[0]);
  time_registers[1] = btoh32(stfp->registers->time[1]);
}


time_t stfp_btime_get_seconds(uint32_t*  time_registers)
{
  return time_registers[1];
}


unsigned int stfp_btime_get_usecs(uint32_t*  time_registers)
{
  return time_registers[0] & 0x000fffff;
}


unsigned int stfp_btime_get_100ns(uint32_t*  time_registers)
{
  return (time_registers[0] & 0x00f00000) >> 20;
}


unsigned int stfp_time_status(uint32_t*  time_registers)
{
  return (time_registers[0] & 0x0f000000) >> 20;
}


const char*  stfp_timing_mode_string(uint8_t  timing_mode)
{
  static const char*  mode_string[7] =
    {
      "Time Code",
      "Free Running",
      "1PPS",
      "Real Time Clock",
      "",
      "",
      "GPS"
    };
  static const char*  empty_string = "";

  if (timing_mode >= (sizeof(mode_string) / sizeof(mode_string[0])))
    return empty_string;

  return mode_string[timing_mode];
}


const char*  stfp_time_format_string(uint8_t  time_format)
{
  static const char*  format_string[7] =
    {
      "BCD",
      "Unix"
    };  
  static const char*  empty_string = "";

  if (time_format >= (sizeof(format_string) / sizeof(format_string[0])))
    return empty_string;

  return format_string[time_format];
}


const char*  stfp_time_code_format_string(uint8_t  format)
{
  static const char*  empty_string  = "";
  static const char*  IRIGA_string  = "IRIG A";
  static const char*  IRIGB_string  = "IRIG B";
  static const char*  IEEE_string   = "IEEE 1344 (1kHz)";
  static const char*  NASA36_string = "NASA36";

  switch (format)
  {
    case 'A' :
      return IRIGA_string;
    case 'B' :
      return IRIGB_string;
    case 'I' :
      return IEEE_string;
    case 'N' :
      return NASA36_string;
    default :
      return empty_string;
  }
}


const char*  stfp_time_code_modulation_string(uint8_t  modulation)
{
  static const char*  empty_string      = "";
  static const char*  amplitude_string  = "amplitude modulated sine wave";
  static const char*  pulse_string      = "pulse code modulation";

  switch (modulation)
  {
    case 'M' :
      return amplitude_string;
    case 'D' :
      return pulse_string;
    default :
      return empty_string;
  }
}


const char*  stfp_selected_clock_source_string(uint8_t  clock_source)
{
  static const char*  empty_string     = "";
  static const char*  internal_string  = "internal 10 MHz oscilator";
  static const char*  external_string  = "external 10 MHz clock";

  switch (clock_source)
  {
    case 'I' :
      return internal_string;
    case 'E' :
      return external_string;
    default :
      return empty_string;
  }
}




struct stfp_device*  new_stfp_device(const char*  base_device_name)
{
  struct stfp_device*  device;
  const char*          config_name = "/config";
  const char*          mem_name    = "/mem.le";
  const char*          intr_name   = "/intr";
  int                  i;

  if (base_device_name == 0)
  {
    printf("base_device_name is NULL in new_stfp_device().\n");
    return 0;
  }

  device = (struct stfp_device*)malloc(sizeof(*device));

  device->config_name = (char*)malloc(strlen(base_device_name)
                                      + strlen(config_name)
                                      + 1);
  strcpy(device->config_name, base_device_name);
  strcat(device->config_name, config_name);

  device->mem_name = (char*)malloc(strlen(base_device_name)
                                   + strlen(mem_name)
                                   + 1);
  strcpy(device->mem_name, base_device_name);
  strcat(device->mem_name, mem_name);

  device->intr_name = (char*)malloc(strlen(base_device_name)
                                   + strlen(intr_name)
                                   + 1);
  strcpy(device->intr_name, base_device_name);
  strcat(device->intr_name, intr_name);

  if ((device->config_handle = open(device->config_name, O_RDONLY)) < 0)
  {
    perror(device->config_name);
    free(device->config_name);
    free(device->mem_name);
    return 0;
  }

  if (ioctl(device->config_handle, PCIIOCGETID, &device->pci_id) < 0)
  {
    perror("ioctl(device->config_handle, PCIIOCGETID,...)");
    free(device->config_name);
    free(device->mem_name);
    free(device->intr_name);
    close(device->config_handle);
    return 0;
  }

  if (BC_VENDOR_ID != (device->pci_id & DEVICE_MASK) ||
      BC635_DEVICE_ID != (((device->pci_id & VENDOR_MASK) >> 16) & 0xffff))
  {
    printf("Device %s does not apear to be a bc635pci board.\n",
           base_device_name);
    free(device->config_name);
    free(device->mem_name);
    free(device->intr_name);
    close(device->config_handle);
  }

  if (ioctl(device->config_handle, PCIIOCGETREV, &device->pci_revision) < 0)
  {
    perror("ioctl(device->config_handle, PCIIOCGETREV,...)");
    free(device->config_name);
    free(device->mem_name);
    free(device->intr_name);
    close(device->config_handle);
    return 0;
  }

  if (ioctl(device->config_handle, PCIIOCGETHTYPE, &device->pci_header) < 0)
  {
    perror("ioctl(device->config_handle, PCIIOCGETHTYPE,...)");
    free(device->config_name);
    free(device->mem_name);
    free(device->intr_name);
    close(device->config_handle);
    return 0;
  }

  for (i = 0; i < 6; i++)
  {
    if (ioctl(device->config_handle, PCIIOCGETBASE(i), &device->bar[i]) < 0)
    {
      perror("ioctl(device->config_handle, PCIIOCGETBASE,...)");
      free(device->config_name);
      free(device->mem_name);
      free(device->intr_name);
      close(device->config_handle);
      return 0;
    }
  }

  if ((device->intr_handle = open(device->intr_name, O_RDWR)) < 0)
  {
    perror(device->intr_name);
    free(device->config_name);
    free(device->mem_name);
    free(device->intr_name);
    close(device->config_handle);
    return 0;
  }

  if ((device->mem_handle = open(device->mem_name, O_RDWR)) < 0)
  {
    perror(device->mem_name);
    free(device->config_name);
    free(device->mem_name);
    free(device->intr_name);
    close(device->config_handle);
    close(device->intr_handle);
    return 0;
  }

  /* The code that follows is a work-around due to a bug in Irix.  The Irix
   * hardware graph should have a device associated with BAR0 for the device
   * registers and another device associated with BAR1 for the dual-port RAM.
   * We should open both above and mmap both here.  Unfortunately the device
   * associated with BAR1 is missing because Irix assumes all BARs are 16k
   * alligned but this one is 4k aligned.  Sgi is fixing the bug.
   *
   * Here we mmap the mem device for the size of the device register area.  Irix
   * will actually map a full page.  We use this fact to get access to the
   * dual-port RAM with an offset from the start of the device registers.
   */

  if ((device->registers = (struct stfp_registers*)mmap(NULL,
                                                        (size_t)0x40,
                                                        (int)(PROT_READ|PROT_WRITE),
                                                        (int) MAP_SHARED,
                                                        device->mem_handle,
                                                        (off_t)0x00))
      == MAP_FAILED)
  {
    printf("Failed mmap() for %s\n", device->mem_name);
    perror("mmap");
    free(device->config_name);
    free(device->mem_name);
    free(device->intr_name);
    close(device->config_handle);
    close(device->intr_handle);
    close(device->mem_handle);
    return 0;
  }

  device->dpram = (uint8_t*)((char*)device->registers + 0x1000);

  {
    volatile uint8_t*  dpram_top;
    int input_area_offset;
    int output_area_offset;
    int gps_area_offset;
    int year_area_offset;

    dpram_top = device->dpram + (dpram_size - 1);

    input_area_offset  = ((int)*dpram_top)       | (((int)*(dpram_top - 1)) << 8);
    output_area_offset = ((int)*(dpram_top - 2)) | (((int)*(dpram_top - 3)) << 8);
    gps_area_offset    = ((int)*(dpram_top - 4)) | (((int)*(dpram_top - 5)) << 8);
    year_area_offset   = ((int)*(dpram_top - 6)) | (((int)*(dpram_top - 7)) << 8);

#if 0
    printf("DPRAM:     0x%x\n", device->dpram);
    printf("DPRAM Top Offset: 0x%x - 1 = 0x%x (%d)\n",
           dpram_size, dpram_size - 1, dpram_size - 1);
    printf("DPRAM Top:     0x%x\n", dpram_top);
    printf("Input Area Offset:  0x%04x\n", input_area_offset);
    printf("Output Area Offset: 0x%04x\n", output_area_offset);
    printf("GPS Area Offset:    0x%04x\n", gps_area_offset);
    printf("Year Area Offset:   0x%04x\n", year_area_offset);
#endif

    device->input_area  = device->dpram + input_area_offset;
    device->output_area = device->dpram + output_area_offset;
    device->gps_area    = device->dpram + gps_area_offset;
    device->year_area   = device->dpram + year_area_offset;
  }

  device->interrupt_stack = malloc(INTERRUPT_STACK_SIZE);

#if CYCLE_COUNTER
  stfp_open_cycle_counter(device);
#endif

  return device;
}


void delete_stfp_device(struct stfp_device**  device_pointer)
{
  struct stfp_device*  device;

  device = *device_pointer;

  free(device->config_name);
  free(device->mem_name);
  free(device->intr_name);
  close(device->config_handle);
  close(device->mem_handle);
  close(device->intr_handle);

#if CYCLE_COUNTER
  close(device->mmem_fd);
#endif

  *device_pointer = NULL;
}


void print_PCI_device_configuration(struct stfp_device*  stfp)
{
  int  i;

  printf(" Card Configuration\n");
  printf("  PCI Device ID: 0x%x\n",
         (((stfp->pci_id & VENDOR_MASK) >> 16) & 0xffff));
  printf("  PCI Vendor ID: 0x%x\n", (stfp->pci_id & DEVICE_MASK));

  for (i = 0; i < 6; i++)
  {
    printf("  PCI Base Address Register %d: 0x%x\n", i, stfp->bar[i]);
  }
}


void print_control_registers(struct stfp_device*  stfp)
{
  time_t    system_time;
  uint32_t  board_time[2];
  time_t    event_time;
  char      system_time_string[30];
  char      board_time_string[30];
  char      event_time_string[30];

  printf(" PCI-BC635  Control Registers Dump \n");
  printf(" ------------------------------------------\n");

  printf(" timereq     -> *meaningless*\n");
  printf(" eventreq    -> *meaningless*\n");
  printf(" unlock      -> *meaningless*\n");
  printf(" reserved    -> *undefined*\n");
  printf(" control     -> 0x%08x\n", stfp->registers->control);
  printf(" ack         -> 0x%08x\n", stfp->registers->ack);
  printf(" mask        -> 0x%08x\n", stfp->registers->mask);
  printf(" intstat     -> 0x%08x\n", stfp->registers->intstat);
  printf(" minstrb     -> 0x%08x\n", stfp->registers->minstrb);
  printf(" majstrb     -> 0x%08x\n", stfp->registers->majstrb);
  printf(" reserved[0] -> *undefined*\n");
  printf(" reserved[1] -> *undefined*\n");
  printf(" time[0]     -> 0x%08x\n", stfp->registers->time[0]);
  stfp_get_time_registers(stfp, board_time);
  strftime(board_time_string, sizeof(board_time_string) - 1,
           "%c", localtime((time_t*)&board_time[1]));
  printf(" time[1]     -> 0x%08x (%s)\n", stfp->registers->time[1], board_time_string);
  printf(" event[0]    -> 0x%08x\n", stfp->registers->event[0]);
  event_time = btoh32(stfp->registers->event[1]);
  strftime(event_time_string, sizeof(event_time_string) - 1,
           "%c", localtime(&event_time));
  printf(" event[1]    -> 0x%08x (%s)\n", stfp->registers->event[1], event_time_string);

  system_time = time(NULL);
  strftime(system_time_string, sizeof(system_time_string) - 1,
           "%c", localtime(&system_time));
  printf(" ---\n system time is 0x%08x (%s)\n", system_time, system_time_string);
}


void print_DPRAM(struct stfp_device*  stfp)
{
  int i; 

  printf(" PCI-BC635 Dual Port RAM  Dump\n");
  printf(" -------------------------------------------------------");
  for (i = 0; i < dpram_size; i += 4)  
  {
    if (i % 0x10 == 0)
      printf ("\n [%03x] : ", i);
    printf("%02x %02x %02x %02x ",
           stfp->dpram[i],
           stfp->dpram[i+1],
           stfp->dpram[i+2],
           stfp->dpram[i+3]);
  }

  printf("\n");
}


void stfp_send_command(struct stfp_device*  stfp,
                       uint8_t              command,
                       const uint8_t*       argument_buffer,
                       int                  nbytes)
{
  stfp->input_area[0] = command;

  if (argument_buffer != NULL && nbytes > 0)
    memmove((void*)(stfp->input_area + 1), argument_buffer, nbytes);

  /* Setting this bit will cause the TFP to zero it so that when it gets set to
   * 1 again we will know the TFP has gotten our command.
   */
  stfp->registers->ack = 0x01000000;
#if TEST
  print_control_registers(stfp);
#endif

  /* Tell the TFP to read our command from the DPRAM Input Area.
   */
  stfp->registers->ack = 0x80000000;
#if TEST
  print_DPRAM(stfp);
#endif

  /* Wait for the TFP to acknowledge the command.
   * Hmmmm... probably should make this time out somehow.
   */
  while (!(stfp->registers->ack & 0x01000000))
    ;
}


int stfp_set_timing_mode(struct stfp_device*  stfp,
                         uint8_t              timing_mode)
{
  uint8_t valid_timing_mode;

  switch (timing_mode)
  {
    case 0x00:  /* Time Code Mode */
    case 0x01:  /* Free Running Mode */
    case 0x02:  /* 1pps Mode */
    case 0x03:  /* Real Time Clock Mode */
    case 0x06:  /* GPS Mode */
      valid_timing_mode = timing_mode;
      break;

    default:
      return 0;
  }
  
  stfp_send_command(stfp, 0x10, &valid_timing_mode, sizeof(valid_timing_mode));
  
  return 1;
}


int stfp_set_time_register_format(struct stfp_device*  stfp,
                                  uint8_t              time_register_format)
{
  uint8_t valid_time_register_format;

  switch (time_register_format)
  {
    case 0x00:
    case 0x01:
      valid_time_register_format = time_register_format;
      break;

    default:
      return 0;
  }
  
  stfp_send_command(stfp, 0x10, &valid_time_register_format,
                    sizeof(valid_time_register_format));
  
  return 1;
}


void  stfp_set_major_time_unix(struct stfp_device*  stfp,
                               time_t               unix_time)
{
  uint32_t  argument = unix_time;
  stfp_send_command(stfp, 0x12, (uint8_t*)&argument, sizeof(argument));
}


int  stfp_set_major_time_bcd(struct stfp_device*  stfp,
                             uint16_t             year,
                             uint16_t             days,
                             uint8_t              hours,
                             uint8_t              minutes,
                             uint8_t              seconds)
{
  uint8_t   argument[7];
  uint8_t*  argp;

  if (year < 1970 || year > 2050 ||
      days > 266 ||
      hours > 23 ||
      minutes > 59 ||
      seconds > 59)
    return 0;

  argp = (uint8_t*)&year;
  argument[0] = argp[0];
  argument[1] = argp[1];

  argp = (uint8_t*)&year;
  argument[2] = argp[0];
  argument[3] = argp[1];

  argument[4] = hours;
  argument[5] = minutes;
  argument[6] = seconds;

  stfp_send_command(stfp, 0x12, argument, sizeof(argument));

  return 1;
}


int  stfp_set_year(struct stfp_device*  stfp,
                   uint16_t             year)
{
  uint8_t   argument[2];
  uint8_t*  argp;

  if (year < 1970 || year > 2050)
    return 0;

  argp = (uint8_t*)&year;
  argument[0] = argp[0];
  argument[1] = argp[1];
  stfp_send_command(stfp, 0x13, argument, sizeof(argument));

  return 1;
}


int  stfp_set_periodic_output(struct stfp_device*  stfp,
                              uint8_t              sync_flag,
                              uint16_t             divider_n1,
                              uint16_t             divider_n2)
{
  uint8_t   argument[5];
  uint8_t*  argp;

  if (sync_flag > 1 ||
      divider_n1 < 2 ||
      divider_n2 < 2)
    return 0;

  argument[0] = sync_flag;

  argp = (uint8_t*)&divider_n1;
  argument[1] = argp[0];
  argument[2] = argp[1];

  argp = (uint8_t*)&divider_n2;
  argument[3] = argp[0];
  argument[4] = argp[1];

  stfp_send_command(stfp, 0x14, argument, sizeof(argument));

  return 1;
}


int stfp_set_input_time_code_format(struct stfp_device*  stfp,
                                    uint8_t              format)
{
  uint8_t valid_format;

  switch (format)
  {
    case 0x41:  /* IRIG A */
    case 0x42:  /* IRIG B */
    case 0x49:  /* IEEE 1244 (1 kHz) */
    case 0x4e:  /* NASA36 */
      valid_format = format;
      break;

    default:
      return 0;
  }
  
  stfp_send_command(stfp, 0x15, &valid_format, sizeof(valid_format));
  
  return 1;
}


int stfp_set_input_time_code_modulation_type(struct stfp_device*  stfp,
                                             uint8_t              modulation)
{
  uint8_t valid_modulation;

  switch (modulation)
  {
    case 0x4D:  /* amplitude modulated sine wave */
    case 0x44:  /* pulse code modulation (DC level shift) */
      valid_modulation = modulation;
      break;

    default:
      return 0;
  }
  
  stfp_send_command(stfp, 0x16, &valid_modulation, sizeof(valid_modulation));
  
  return 1;
}


void  stfp_set_propagation_delay_compensation(struct stfp_device*  stfp,
                                              int32_t              offset)
{
  uint8_t   argument[4];
  uint8_t*  argp;

  argp = (uint8_t*)&offset;
  argument[0] = argp[0];
  argument[1] = argp[1];
  argument[2] = argp[2];
  argument[3] = argp[3];

  stfp_send_command(stfp, 0x17, argument, sizeof(argument));
}


int stfp_request_data_timing_mode(struct stfp_device*  stfp,
                                  uint8_t*             timing_mode)
{
  const uint8_t  request = 0x10;

  stfp_send_command(stfp, 0x19, &request, sizeof(request));

  if (stfp->output_area[0] != request)
    return 0;

  if (timing_mode != NULL)
    timing_mode[0] = stfp->output_area[1];

  return 1;
}


int stfp_request_data_time_format(struct stfp_device*  stfp,
                                  uint8_t*             time_format)
{
  const uint8_t  request = 0x11;

  stfp_send_command(stfp, 0x19, &request, sizeof(request));

  if (stfp->output_area[0] != request)
    return 0;

  if (time_format != NULL)
    time_format[0] = stfp->output_area[1];

  return 1;
}


int stfp_request_data_current_year(struct stfp_device*  stfp,
                                   uint16_t*            year)
{
  const uint8_t  request = 0x13;

  stfp_send_command(stfp, 0x19, &request, sizeof(request));

  if (stfp->output_area[0] != request)
    return 0;

  if (year != NULL)
    memmove(year, (void*)(stfp->output_area + 1), sizeof(*year));

  return 1;
}


int stfp_request_data_periodic_output(struct stfp_device*  stfp,
                                      uint8_t*             synch,
                                      uint16_t*            divider_n1,
                                      uint16_t*            divider_n2)
{
  const uint8_t  request = 0x14;

  stfp_send_command(stfp, 0x19, &request, sizeof(request));

  if (stfp->output_area[0] != request)
    return 0;

  if (synch != NULL)
    synch[0] = stfp->output_area[1];

  if (divider_n1 != NULL)
    memmove(divider_n1, (void*)(stfp->output_area + 2), sizeof(*divider_n1));

  if (divider_n2 != NULL)
    memmove(divider_n2, (void*)(stfp->output_area + 4), sizeof(*divider_n2));

  return 1;
}


int stfp_request_data_time_code_format(struct stfp_device*  stfp,
                                       uint8_t*             format)
{
  const uint8_t  request = 0x15;

  stfp_send_command(stfp, 0x19, &request, sizeof(request));

  if (stfp->output_area[0] != request)
    return 0;

  if (format != NULL)
    format[0] = stfp->output_area[1];

  return 1;
}


int stfp_request_data_time_code_modulation(struct stfp_device*  stfp,
                                           uint8_t*             modulation)
{
  const uint8_t  request = 0x16;

  stfp_send_command(stfp, 0x19, &request, sizeof(request));

  if (stfp->output_area[0] != request)
    return 0;

  if (modulation != NULL)
    modulation[0] = stfp->output_area[1];

  return 1;
}


int stfp_request_data_timing_offset(struct stfp_device*  stfp,
                                    int32_t*             offset)
{
  const uint8_t  request = 0x17;

  stfp_send_command(stfp, 0x19, &request, sizeof(request));

  if (stfp->output_area[0] != request)
    return 0;

  if (offset != NULL)
    memmove(offset, (void*)(stfp->output_area + 1), sizeof(*offset));

  return 1;
}


int stfp_request_data_selected_clock_source(struct stfp_device*  stfp,
                                            uint8_t*             clock_source)
{
  const uint8_t  request = 0x20;

  stfp_send_command(stfp, 0x19, &request, sizeof(request));

  if (stfp->output_area[0] != request)
    return 0;

  if (clock_source != NULL)
    clock_source[0] = stfp->output_area[1];

  return 1;
}


int stfp_request_data_firmware_part_number(struct stfp_device*  stfp,
                                           char*                firmware_part)
{
  const uint8_t  request = 0x4f;

  stfp_send_command(stfp, 0x19, &request, sizeof(request));

  if (stfp->output_area[0] != request)
    return 0;

  if (firmware_part != NULL)
  {
    firmware_part[0]  = stfp->output_area[1];
    firmware_part[1]  = stfp->output_area[2];
    firmware_part[2]  = stfp->output_area[3];
    firmware_part[3]  = stfp->output_area[4];
    firmware_part[4]  = stfp->output_area[5];
    firmware_part[5]  = stfp->output_area[6];
    firmware_part[6]  = stfp->output_area[7];
    firmware_part[7]  = stfp->output_area[8];
    firmware_part[8]  = stfp->output_area[9];
    firmware_part[9]  = stfp->output_area[10];
    firmware_part[10] = stfp->output_area[11];
    firmware_part[11] = 0;
  }

  return 1;
}


int stfp_request_data_assembly_part_number(struct stfp_device*  stfp,
                                           uint16_t*            assembly_part)
{
  const uint8_t  request = 0xf4;

  stfp_send_command(stfp, 0x19, &request, sizeof(request));

  if (stfp->output_area[0] != request)
    return 0;

  if (assembly_part != NULL)
    memmove(assembly_part, (void*)(stfp->output_area + 1), sizeof(*assembly_part));

  return 1;
}


int stfp_request_data_fab_part_number(struct stfp_device*  stfp,
                                      uint16_t*            fab_part)
{
  const uint8_t  request = 0xf5;

  stfp_send_command(stfp, 0x19, &request, sizeof(request));

  if (stfp->output_area[0] != request)
    return 0;

  if (fab_part != NULL)
    memmove(fab_part, (void*)(stfp->output_area + 1), sizeof(*fab_part));

  return 1;
}


int stfp_request_data_model_identification(struct stfp_device*  stfp,
                                           char*                model)
{
  const uint8_t  request = 0xf6;

  stfp_send_command(stfp, 0x19, &request, sizeof(request));

  if (stfp->output_area[0] != request)
    return 0;

  if (model != NULL)
  {
    model[0] = stfp->output_area[1];
    model[1] = stfp->output_area[2];
    model[2] = stfp->output_area[3];
    model[3] = stfp->output_area[4];
    model[4] = stfp->output_area[5];
    model[5] = stfp->output_area[6];
    model[6] = stfp->output_area[7];
    model[7] = stfp->output_area[8];
    model[8] = stfp->output_area[9];
    model[9] = 0;
  }

  return 1;
}


int stfp_request_data_serial_number(struct stfp_device*  stfp,
                                    uint32_t*            serial_number)
{
  const uint8_t  request = 0xf7;

  stfp_send_command(stfp, 0x19, &request, sizeof(request));

  if (stfp->output_area[0] != request)
    return 0;

  if (serial_number != NULL)
    memmove(serial_number, (void*)(stfp->output_area + 1), sizeof(*serial_number));

  return 1;
}


void stfp_software_reset(struct stfp_device*  stfp)
{
  stfp_send_command(stfp, 0x1a, NULL, 0);
}


int stfp_select_clock_source(struct stfp_device*  stfp,
                             uint8_t              clock_source)
{
  uint8_t valid_clock_source;

  switch (clock_source)
  {
    case 0x49:  /* Internal 10 MHz Oscilator */
    case 0x44:  /* External 10 MHz Clock */
      valid_clock_source = clock_source;
      break;

    default:
      return 0;
  }
  
  stfp_send_command(stfp, 0x20, &valid_clock_source, sizeof(valid_clock_source));
  
  return 1;
}


int stfp_control_jam_sync(struct stfp_device*  stfp,
                          uint8_t              sync)
{
  uint8_t valid_sync;

  switch (sync)
  {
    case 0x00:
    case 0x01:
      valid_sync = sync;
      break;

    default:
      return 0;
  }
  
  stfp_send_command(stfp, 0x21, &valid_sync, sizeof(valid_sync));
  
  return 1;
}


void stfp_force_jam_sync(struct stfp_device*  stfp)
{
  stfp_send_command(stfp, 0x22, NULL, 0);
}


/* Code for this command was found in the original sample source code.
 * However, this command is not documented in the TFP User's Guide
 * so I don't know if this actually does anything.
 */
void stfp_control_clock_disciplining(struct stfp_device*  stfp,
                                     uint8_t              enable)
{
  uint8_t discipline;

  if (enable)
    discipline = 1;
  else
    discipline = 0;

  stfp_send_command(stfp, 0x23, &discipline, sizeof(discipline));
}


void stfp_load_da_converter(struct stfp_device*  stfp,
                            uint16_t             value)
{
  stfp_send_command(stfp, 0x24, (uint8_t*)&value, sizeof(value));
}


void stfp_set_disciplining_gain(struct stfp_device*  stfp,
                                int16_t              gain)
{
  stfp_send_command(stfp, 0x25, (uint8_t*)&gain, sizeof(gain));
}


int stfp_register_interrupt_function(struct stfp_device*  stfp,
                                     void (*interrupt_function)(void*),
                                     void*  interrupt_function_argument)
{
  stfp->ULI_id = ULI_register_pci(stfp->intr_handle,  /* the pci interrupt device */
                                  interrupt_function, /* the handler function pointer */
                                  interrupt_function_argument, /* the argument to the handler */
                                  0,                  /* the # of semaphores needed  */
                                  stfp->interrupt_stack,  /* the stack to use */
                                  INTERRUPT_STACK_SIZE,   /* the stack size to use */
                                  0x01);              /* PCI interrrupt line */

  if (stfp->ULI_id == 0)
  {
    return 0;
  }

  return 1;
}
